#
# Copyright 2013 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
This module represents the N'Ko language.

.. seealso:: :wp:`N'Ko_language`
"""

import re

from translate.lang import common


def reverse_quotes(text):
    def convertquotation(match):
        return f"”{match.group(1)}“"

    return re.sub(r"“([^”]+)”", convertquotation, text)


class nqo(common.Common):
    """This class represents N'Ko."""

    listseperator = "߸ "

    puncdict = {
        ",": "߸",
        ";": "؛",
        "?": "؟",
        "!": "߹",
    }

    ignoretests = {
        "all": ["acronyms", "simplecaps", "startcaps"],
    }

    @classmethod
    def punctranslate(cls, text):
        text = super().punctranslate(text)
        return reverse_quotes(text)



# po2tiki unit tests
# Author: Wil Clouser <wclouser@mozilla.com>
# Date: 2008-12-01
from io import BytesIO

from translate.convert import po2tiki

from . import test_convert


class TestPo2Tiki:
    ConverterClass = po2tiki.po2tiki

    def _convert(self, input_string, template_string=None, success_expected=True):
        """Helper that converts to target format without using files."""
        input_file = BytesIO(input_string.encode())
        output_file = BytesIO()
        template_file = None
        if template_string:
            template_file = BytesIO(template_string.encode())
        expected_result = 1 if success_expected else 0
        converter = self.ConverterClass(input_file, output_file, template_file)
        assert converter.run() == expected_result
        return converter.target_store, output_file

    def _convert_to_string(self, *args, **kwargs):
        """Helper that converts to target format string without using files."""
        return self._convert(*args, **kwargs)[1].getvalue().decode("utf-8")

    def test_convert_empty(self):
        """Check converting empty file returns no output."""
        assert self._convert_to_string("", success_expected=False) == ""

    def test_convert(self):
        """Check converting simple file."""
        input_string = """
#: translated
msgid "zero_source"
msgstr "zero_target"

#: unused
msgid "one_source"
msgstr "one_target"
"""
        output = self._convert_to_string(input_string)
        assert '"one_source" => "one_target",' in output
        assert '"zero_source" => "zero_target",' in output

    def test_convert_marked_untranslated(self):
        """Check convert marked as untranslated keeps translation."""
        input_string = """
#: untranslated
msgid "Do not translate"
msgstr "It is translated"
"""
        output = self._convert_to_string(input_string)
        assert '"Do not translate" => "It is translated",' in output


class TestPo2TikiCommand(test_convert.TestConvertCommand, TestPo2Tiki):
    """Tests running actual po2tiki commands on files."""

    convertmodule = po2tiki
    defaultoptions = {}



#
# Copyright 2007 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
This module represents the Spanish language.

.. note:: As it only has special case code for initial inverted punctuation,
   it could also be used for Asturian, Galician, or Catalan.
"""

from translate.lang import common


class es(common.Common):
    """This class represents Spanish."""

    @classmethod
    def punctranslate(cls, text):
        """Implement some extra features for inverted punctuation."""
        text = super().punctranslate(text)
        # If the first sentence ends with ? or !, prepend inverted ¿ or ¡
        firstmatch = cls.sentencere.match(text)
        # only one sentence (if any) - use entire string
        first = text if firstmatch is None else firstmatch.group()
        # remove trailing whitespace
        first = first.strip()
        # protect against incorrectly handling an empty string
        if not first:
            return text
        if first[-1] == "?":
            text = "¿" + text
        elif first[-1] == "!":
            text = "¡" + text
        return text



#
# Copyright 2015 Zuza Software Foundation
# Copyright 2015 Sarah Hale
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Convert .Net Resource (.resx) to Gettext PO localisation files.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/resx2po.html
for examples and usage instructions.
"""

import logging

from translate.storage import po

logger = logging.getLogger(__name__)


class resx2po:
    """Convert a RESX file to a PO file for handling translation."""

    def convert_store(self, input_store, duplicatestyle="msgctxt"):
        """Converts a RESX file to a PO file."""
        output_store = po.pofile()
        output_header = output_store.init_headers(
            charset="UTF-8", encoding="8bit", x_accelerator_marker="&"
        )
        output_header.addnote(f"extracted from {input_store.filename}", "developer")
        for input_unit in input_store.units:
            if input_unit.istranslatable():
                output_unit = self.convert_unit(input_unit, "developer")
                if output_unit is not None:
                    # Split out translator & dev comments before adding them
                    self.split_comments(output_unit, output_unit)

                    output_store.addunit(output_unit)

        output_store.removeduplicates(duplicatestyle)
        return output_store

    def merge_store(
        self, template_store, input_store, blankmsgstr=False, duplicatestyle="msgctxt"
    ):
        """Converts two RESX files to a PO file."""
        output_store = po.pofile()
        output_header = output_store.init_headers(
            charset="UTF-8", encoding="8bit", x_accelerator_marker="&"
        )
        output_header.addnote(
            f"extracted from {template_store.filename}, {input_store.filename}",
            "developer",
        )

        input_store.makeindex()
        for template_unit in template_store.units:
            origpo = self.convert_unit(template_unit, "developer")

            # try and find a translation of the same name...
            template_unit_name = "".join(template_unit.getlocations())
            if template_unit_name in input_store.locationindex:
                translatedresx = input_store.locationindex[template_unit_name]
                translatedpo = self.convert_unit(translatedresx, "translator")
            else:
                translatedpo = None

            # if we have a valid po unit, get the translation and add it...
            if origpo is not None:
                if translatedpo is not None and not blankmsgstr:
                    origpo.target = translatedpo.source

                    # Split out translator & dev comments before adding them
                    self.split_comments(origpo, translatedpo)

                output_store.addunit(origpo)

            elif translatedpo is not None:
                logger.error("Error converting original RESX definition %s", origpo)

        output_store.removeduplicates(duplicatestyle)
        return output_store

    @staticmethod
    def split_comments(origpo, translatedpo):
        autocomments = translatedpo.getnotes("developer")
        if autocomments:
            devcomment, transcomment = autocomments.partition("[Translator Comment: ")[
                ::2
            ]
            if transcomment:
                origpo.addnote(transcomment.replace("]", ""), origin="translator")
            if devcomment:
                origpo.addnote(devcomment.strip(), origin="developer", position="merge")

    @staticmethod
    def convert_unit(input_unit, commenttype):
        """
        Converts a RESX unit to a PO unit
        @return: None if empty or not for translation.
        """
        if input_unit is None:
            return None

        output_unit = po.pounit(encoding="UTF-8")
        output_unit.addlocation(input_unit.getid())
        output_unit.source = input_unit.source
        output_unit.addnote(input_unit.getnotes("developer"), "developer")
        output_unit.target = ""

        return output_unit


def convert_resx(
    input_file,
    output_file,
    template_file,
    pot=False,
    duplicatestyle="msgctxt",
    filter=None,
):
    from translate.storage import resx

    input_store = resx.RESXFile(input_file)
    convertor = resx2po()
    if template_file is None:
        output_store = convertor.convert_store(
            input_store, duplicatestyle=duplicatestyle
        )
    else:
        template_store = resx.RESXFile(template_file)
        output_store = convertor.merge_store(
            template_store, input_store, blankmsgstr=pot, duplicatestyle=duplicatestyle
        )
    if output_store.isempty():
        return 0
    output_store.serialize(output_file)
    return 1


def main(argv=None):
    from translate.convert import convert

    formats = {
        "resx": ("po", convert_resx),
        ("resx", "resx"): ("po", convert_resx),
    }
    parser = convert.ConvertOptionParser(
        formats, usetemplates=True, usepots=True, description=__doc__
    )
    parser.add_option(
        "",
        "--filter",
        dest="filter",
        default=None,
        help="leaves to extract e.g. 'name,desc': (default: extract everything)",
        metavar="FILTER",
    )
    parser.add_duplicates_option()
    parser.passthrough.append("pot")
    parser.passthrough.append("filter")
    parser.run(argv)


if __name__ == "__main__":
    main()



#
# Copyright 2004-2014 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.


# Copied from git commit 96b9f1419453d8079dd1631c329f04d6e005baae from
# git://hforge.org/itools.git
config_uri = "urn:oasis:names:tc:opendocument:xmlns:config:1.0"
dc_uri = "http://purl.org/dc/elements/1.1/"
form_uri = "urn:oasis:names:tc:opendocument:xmlns:form:1.0"
meta_uri = "urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
number_uri = "urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0"
office_uri = "urn:oasis:names:tc:opendocument:xmlns:office:1.0"
presentation_uri = "urn:oasis:names:tc:opendocument:xmlns:presentation:1.0"
text_uri = "urn:oasis:names:tc:opendocument:xmlns:text:1.0"
svg_uri = "urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0"


inline_elements = [
    (text_uri, "page-count"),
    (text_uri, "page-number"),
    (text_uri, "a"),
    (text_uri, "line-break"),
    (text_uri, "ruby-base"),
    (text_uri, "s"),
    (text_uri, "span"),
    (text_uri, "tab"),
]


no_translate_content_elements = [  # It is named skip_content_elements in itools.
    # Config
    (config_uri, "config-item"),
    # Dublin core
    (dc_uri, "creator"),
    (dc_uri, "date"),
    # (dc_uri, 'description'),
    (dc_uri, "language"),
    # (dc_uri, 'subject'),
    # (dc_uri, 'title'),
    # Form
    (form_uri, "item"),
    (form_uri, "option"),
    # Meta
    (meta_uri, "creation-date"),
    (meta_uri, "date-string"),
    (meta_uri, "editing-cycles"),
    (meta_uri, "editing-duration"),
    (meta_uri, "generator"),
    (meta_uri, "initial-creator"),
    # (meta_uri, 'keyword'),
    (meta_uri, "printed-by"),
    (meta_uri, "print-date"),
    (meta_uri, "user-defined"),
    # Number
    (number_uri, "currency-symbol"),
    (number_uri, "embedded-text"),
    (number_uri, "text"),
    # Office
    (office_uri, "binary-data"),
    # Presentation
    (presentation_uri, "date-time-decl"),
    # (presentation_uri, 'footer-decl'),
    # (presentation_uri, 'header-decl'),
    # Text
    (text_uri, "author-initials"),
    (text_uri, "author-name"),
    # XXX (text_uri, 'bibliography-mark'),
    (text_uri, "bookmark-ref"),
    # (text_uri, 'chapter'),
    (text_uri, "character-count"),
    # (text_uri, 'conditional-text'),
    (text_uri, "creation-date"),
    (text_uri, "creation-time"),
    (text_uri, "creator"),
    (text_uri, "date"),
    (text_uri, "dde-connection"),
    # (text_uri, 'description'),
    (text_uri, "editing-cycles"),
    (text_uri, "editing-duration"),
    (text_uri, "expression"),
    (text_uri, "file-name"),
    # (text_uri, 'hidden-paragraph'),
    # (text_uri, 'hidden-text'),
    (text_uri, "image-count"),
    # (text_uri, 'index-entry-span'),
    (text_uri, "index-title-template"),
    (text_uri, "initial-creator"),
    # (text_uri, 'keywords'),
    (text_uri, "linenumbering-separator"),
    (text_uri, "measure"),
    (text_uri, "modification-date"),
    (text_uri, "modification-time"),
    # (text_uri, 'note-citation'),
    # (text_uri, 'note-continuation-notice-backward'),
    # (text_uri, 'note-continuation-notice-forward'),
    (text_uri, "note-ref"),
    (text_uri, "number"),
    (text_uri, "object-count"),
    (text_uri, "page-continuation"),
    (text_uri, "page-count"),
    (text_uri, "page-number"),
    (text_uri, "page-variable-get"),
    (text_uri, "page-variable-set"),
    (text_uri, "paragraph-count"),
    # (text_uri, 'placeholder'),
    (text_uri, "print-date"),
    (text_uri, "print-time"),
    (text_uri, "printed-by"),
    (text_uri, "reference-ref"),
    # (text_uri, 'ruby-text'),
    (text_uri, "script"),
    (text_uri, "sender-city"),
    (text_uri, "sender-company"),
    (text_uri, "sender-country"),
    (text_uri, "sender-email"),
    (text_uri, "sender-fax"),
    (text_uri, "sender-firstname"),
    (text_uri, "sender-initials"),
    (text_uri, "sender-lastname"),
    (text_uri, "sender-phone-private"),
    (text_uri, "sender-phone-work"),
    # (text_uri, 'sender-position'),
    (text_uri, "sender-postal-code"),
    (text_uri, "sender-state-or-province"),
    (text_uri, "sender-street"),
    # (text_uri, 'sender-title'),
    (text_uri, "sequence"),
    (text_uri, "sequence-ref"),
    (text_uri, "sheet-name"),
    # (text_uri, 'subject'),
    (text_uri, "table-count"),
    (text_uri, "table-formula"),
    (text_uri, "template-name"),
    (text_uri, "text-input"),
    (text_uri, "time"),
    # (text_uri, 'title'),
    (text_uri, "user-defined"),
    (text_uri, "user-field-get"),
    (text_uri, "user-field-input"),
    (text_uri, "variable-get"),
    (text_uri, "variable-input"),
    (text_uri, "variable-set"),
    (text_uri, "word-count"),
    # SVG
    # (svg_uri, 'title'),
    # (svg_uri, 'desc')
    # From translate
    (text_uri, "tracked-changes"),
]



from ast import literal_eval

from translate.convert import web2py2po
from translate.storage import po

from ..storage.test_base import first_translatable, headerless_len


class TestWEB2PY2PO:
    @staticmethod
    def web2py2po(web2py_source):
        """Helper that converts po source to web2py source without requiring files."""
        input_web2py = literal_eval(web2py_source)
        new_pofile = po.pofile()
        convertor = web2py2po.web2py2po(new_pofile)
        return convertor.convertstore(input_web2py)

    @staticmethod
    def singleelement(storage):
        """Checks that the pofile contains a single non-header element, and returns it."""
        assert headerless_len(storage.units) == 1
        return first_translatable(storage)

    def test_basic(self):
        """Test a basic web2py to po conversion."""
        input_web2py = """# -*- coding: utf-8 -*-
{
'A simple string': 'Du texte simple',
}
"""

        po_out = self.web2py2po(input_web2py)
        pounit = self.singleelement(po_out)
        assert pounit.source == "A simple string"
        assert pounit.target == "Du texte simple"

    def test_unicode(self):
        """Test a web2py to po conversion with unicode."""
        input_web2py = """# -*- coding: utf-8 -*-
{
'Foobar': 'Fúbär',
}
"""

        po_out = self.web2py2po(input_web2py)
        pounit = self.singleelement(po_out)
        assert pounit.source == "Foobar"
        assert pounit.target == "Fúbär"

    def test_markmin(self):
        """Test removal of @markmin in po to web2py conversion."""
        input_web2py = """# -*- coding: utf-8 -*-
{
'@markmin\\x01Hello **world**!': 'Hello **world**!',
}
"""

        po_out = self.web2py2po(input_web2py)
        pounit = self.singleelement(po_out)
        assert pounit.source == "@markmin\x01Hello **world**!"
        assert pounit.target == ""



import subprocess
import sys
from io import BytesIO

from pytest import CaptureFixture, mark, param

from translate.storage import po
from translate.tools import pocount

# For now test files left in the old places, but it's better to move
# them somwhere like tests/data.
_po_csv, _po_file, _po_fuzzy = test_files = [
    "tests/cli/data/test_pocount_po_csv/one.po",
    "tests/cli/data/test_pocount_po_file/one.po",
    "tests/cli/data/test_pocount_po_fuzzy/one.po",
]
_xliff_states_yes = "tests/cli/data/test_pocount_xliff_states_yes/states.xlf"
_xliff_states_no = "tests/cli/data/test_pocount_xliff_states_no/states.xlf"


class TestCount:
    @staticmethod
    def count(source, expectedsource, target=None, expectedtarget=None):
        """Simple helper to check the respective word counts."""
        poelement = po.pounit(source)
        if target is not None:
            poelement.target = target
        wordssource, wordstarget = pocount.wordsinunit(poelement)
        print(
            'Source (expected=%d; actual=%d): "%s"'
            % (expectedsource, wordssource, source)
        )
        assert wordssource == expectedsource
        if target is not None:
            print(
                'Target (expected=%d; actual=%d): "%s"'
                % (expectedtarget, wordstarget, target)
            )
            assert wordstarget == expectedtarget

    def test_simple_count_zero(self):
        """No content."""
        self.count("", 0)

    def test_simple_count_one(self):
        """Simplest one word count."""
        self.count("One", 1)

    def test_simple_count_two(self):
        """Simplest one word count."""
        self.count("One two", 2)

    def test_punctuation_divides_words(self):
        """Test that we break words when there is punctuation."""
        self.count("One. Two", 2)
        self.count("One.Two", 2)

    def test_xml_tags(self):
        """Test that we do not count XML tags as words."""
        # <br> is a word break
        self.count("A word<br>Another word", 4)
        self.count("A word<br/>Another word", 4)
        self.count("A word<br />Another word", 4)
        # \n is a word break
        self.count("<p>A word</p>\n<p>Another word</p>", 4)
        # Not really an XML tag
        self.count("<no label>", 2)

    def test_newlines(self):
        """Test to see that newlines divide words."""
        # newlines break words
        self.count("A word.\nAnother word", 4)
        self.count(r"A word.\\n\nAnother word", 4)

    def test_variables_are_words(self):
        """Test that we count variables as words."""
        self.count("%PROGRAMNAME %PROGRAM% %s $file $1", 5)

    def test_plurals(self):
        """Test that we can handle plural PO elements."""
        # #: gdk-pixbuf/gdk-pixdata.c:430
        # #, c-format
        # msgid "failed to allocate image buffer of %u byte"
        # msgid_plural "failed to allocate image buffer of %u bytes"
        # msgstr[0] "e paletšwe go hwetša sešireletši sa seswantšho sa paete ya %u"
        # msgstr[1] "e paletšwe go hwetša sešireletši sa seswantšho sa dipaete tša %u"

    @mark.xfail(reason="Support commented out pending removal")
    def test_plurals_kde(self):
        """Test that we correcly count old style KDE plurals."""
        self.count("_n: Singular\\n\nPlural", 2, "Een\\n\ntwee\\n\ndrie", 3)

    def test_msgid_blank(self):
        """Counts a message id."""
        self.count("   ", 0)

    # Counting strings
    #  We need to check how we count strings also and if we call it translated or untranslated
    # ie an all spaces msgid should be translated if there are spaces in the msgstr

    # Make sure we don't count obsolete messages

    # Do we correctly identify a translated yet blank message?

    # Need to test that we can differentiate between fuzzy, translated and untranslated


class TestPOCount:
    """This only tests the old (memory-based) pocount method."""

    inputdata = rb"""
msgid "translated unit"
msgstr "translated unit"

#, fuzzy
msgid "fuzzy unit"
msgstr "fuzzy unit"

# untranslated
msgid "untranslated unit"
msgstr ""

# obsolete
#~ msgid "obsolete translated unit"
#~ msgstr "obsolete translated unit"

#, fuzzy
#~ msgid "obsolete fuzzy unit"
#~ msgstr "obsolete fuzzy unit"

# untranslated
#~ msgid "obsolete untranslated unit"
#~ msgstr ""
"""

    def test_translated(self):
        pofile = BytesIO(self.inputdata)
        stats = pocount.calcstats(pofile)
        assert stats["translated"] == 1

    def test_fuzzy(self):
        pofile = BytesIO(self.inputdata)
        stats = pocount.calcstats(pofile)
        assert stats["fuzzy"] == 1

    def test_untranslated(self):
        pofile = BytesIO(self.inputdata)
        stats = pocount.calcstats(pofile)
        assert stats["untranslated"] == 1

    def test_total(self):
        pofile = BytesIO(self.inputdata)
        stats = pocount.calcstats(pofile)
        assert stats["total"] == 3

    def test_translatedsourcewords(self):
        pofile = BytesIO(self.inputdata)
        stats = pocount.calcstats(pofile)
        assert stats["translatedsourcewords"] == 2

    def test_fuzzysourcewords(self):
        pofile = BytesIO(self.inputdata)
        stats = pocount.calcstats(pofile)
        assert stats["fuzzysourcewords"] == 2

    def test_untranslatedsourcewords(self):
        pofile = BytesIO(self.inputdata)
        stats = pocount.calcstats(pofile)
        assert stats["untranslatedsourcewords"] == 2

    def test_totalsourcewords(self):
        pofile = BytesIO(self.inputdata)
        stats = pocount.calcstats(pofile)
        assert stats["totalsourcewords"] == 6


@mark.parametrize("style", ["csv", "full", "short-strings", "short-words"])
@mark.parametrize("incomplete", [True, False], ids=lambda v: f"incomplete={v}")
@mark.parametrize("no_color", [True, False], ids=lambda v: f"no-color={v}")
def test_output(style, incomplete, no_color, capsys: CaptureFixture[str], snapshot):
    opts = [f"--{style}"]
    if incomplete:
        opts.append("--incomplete")
    if no_color:
        opts.append("--no-color")

    pocount.main([*opts, *test_files])
    stdout = capsys.readouterr()[0]

    assert stdout == snapshot


@mark.parametrize(
    "opts",
    [
        param([_po_file, "--no-color"], id="po-file"),
        param([_po_fuzzy, "--no-color"], id="po-file-fuzzy"),
        param([_po_csv, "--no-color", "--csv"], id="po-file-csv"),
        param([_xliff_states_yes, "--no-color"], id="xliff-states-yes"),
        param([_xliff_states_no, "--no-color"], id="xliff-states-no"),
    ],
)
def test_cases(opts, capsys: CaptureFixture[str], snapshot):
    pocount.main(opts)

    result = capsys.readouterr()

    assert result == snapshot


@mark.parametrize(
    ("opts", "expected"),
    [
        param(
            ["--csv", "--short"],
            "argument --short: not allowed with argument --csv",
            id="mutually-exclusive",
        ),
        param(
            ["missing.po"],
            "cannot process missing.po: does not exist",
            id="missing-file",
        ),
        param([], "the following arguments are required: files", id="no-args"),
    ],
)
def test_error_cases(opts, expected):
    # We're using special case for this, to produce correct output.
    # Also, using partial matching instead of snapshots, becouse mac-os argparse
    # output is slightly different.
    result = subprocess.run(
        [sys.executable, pocount.__file__, *opts],
        capture_output=True,
        text=True,
    )

    assert expected in result.stderr



#
# Copyright 2002-2007 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Classes that hold units of .po files (pounit) or entire files (pofile).

Gettext-style .po (or .pot) files are used in translations for KDE, GNOME and
many other projects.

This uses libgettextpo from the gettext package. Any version before 0.17 will
at least cause some subtle bugs or may not work at all. Developers might want
to have a look at gettext-tools/libgettextpo/gettext-po.h from the gettext
package for the public API of the library.
"""

import ctypes.util
import logging
import os
import re
import sys
import tempfile
import threading
from ctypes import (
    CFUNCTYPE,
    POINTER,
    Structure,
    c_char_p,
    c_int,
    c_size_t,
    c_uint,
    cdll,
)

from translate.misc.multistring import multistring
from translate.storage import base, pocommon, pypo

logger = logging.getLogger(__name__)

lsep = " "
"""Separator for #: entries"""

STRING = c_char_p


# Structures
class po_message(Structure):
    pass


class po_file(Structure):
    pass


class po_filepos(Structure):
    pass


class po_iterator(Structure):
    pass


po_message_t = POINTER(po_message)
"""A po_message_t represents a message in a PO file."""

po_file_t = POINTER(po_file)
"""A po_file_t represents a PO file."""

po_filepos_t = POINTER(po_filepos)
"""A po_filepos_t represents the position in a PO file."""

po_iterator_t = POINTER(po_iterator)
"""A po_iterator_t represents an iterator through a PO file."""


# Function prototypes
xerror_prototype = CFUNCTYPE(
    None, c_int, po_message_t, STRING, c_uint, c_uint, c_int, STRING
)
xerror2_prototype = CFUNCTYPE(
    None,
    c_int,
    po_message_t,
    STRING,
    c_uint,
    c_uint,
    c_int,
    STRING,
    po_message_t,
    STRING,
    c_uint,
    c_uint,
    c_int,
    STRING,
)


# Structures (error handler)
class po_xerror_handler(Structure):
    _fields_ = [("xerror", xerror_prototype), ("xerror2", xerror2_prototype)]


class po_error_handler(Structure):
    _fields_ = [
        ("error", CFUNCTYPE(None, c_int, c_int, STRING)),
        ("error_at_line", CFUNCTYPE(None, c_int, c_int, STRING, c_uint, STRING)),
        ("multiline_warning", CFUNCTYPE(None, STRING, STRING)),
        ("multiline_error", CFUNCTYPE(None, STRING, STRING)),
    ]


xerror_storage = threading.local()

ignored_erorrs = {
    # TODO: this is probably bug somewhere in cpo, but
    # it used to be silently ignored before the exceptions
    # were raised, so it is left to fixing separately
    "invalid multibyte sequence",
    # Duplicate messages are allowed
    "duplicate message definition",
}


def trigger_exception(severity, filename, lineno, column, message_text):
    # Severity 0 is warning, severity 1 error, severity 2 critical
    if severity >= 1 and message_text not in ignored_erorrs:
        if filename:
            detail = f"{filename}:{lineno}:{column}: {message_text}"
        else:
            detail = message_text
        xerror_storage.exception = ValueError(detail)


# Callback functions for po_xerror_handler
def xerror_cb(severity, message, filename, lineno, column, multiline_p, message_text):
    message_text = message_text.decode()
    if filename:
        filename = filename.decode()
    logger.error(
        "xerror_cb %s %s %s %s %s %s %s",
        severity,
        message,
        filename,
        lineno,
        column,
        multiline_p,
        message_text,
    )
    trigger_exception(severity, filename, lineno, column, message_text)


def xerror2_cb(
    severity,
    message1,
    filename1,
    lineno1,
    column1,
    multiline_p1,
    message_text1,
    message2,
    filename2,
    lineno2,
    column2,
    multiline_p2,
    message_text2,
):
    message_text1 = message_text1.decode()
    message_text2 = message_text2.decode()
    if filename1:
        filename1 = filename1.decode()
    if filename2:
        filename2 = filename2.decode()
    logger.error(
        "xerror2_cb %s %s %s %s %s %s %s %s %s %s %s %s",
        severity,
        message1,
        filename1,
        lineno1,
        column1,
        multiline_p1,
        message_text1,
        filename2,
        lineno2,
        column2,
        multiline_p2,
        message_text2,
    )
    trigger_exception(severity, filename1, lineno1, column1, message_text1)


# Setup return and parameter types
# See also http://git.savannah.gnu.org/cgit/gettext.git/tree/gettext-tools/libgettextpo/gettext-po.in.h
def setup_call_types(gpo):
    # File access
    gpo.po_file_create.restype = po_file_t
    gpo.po_file_read_v3.argtypes = [STRING, POINTER(po_xerror_handler)]
    gpo.po_file_read_v3.restype = po_file_t
    gpo.po_file_write_v2.argtypes = [po_file_t, STRING, POINTER(po_xerror_handler)]
    gpo.po_file_write_v2.restype = po_file_t
    gpo.po_file_free.argtypes = [po_file_t]

    # Header
    gpo.po_file_domain_header.argtypes = [po_file_t, STRING]
    gpo.po_file_domain_header.restype = STRING
    gpo.po_header_field.argtypes = [STRING, STRING]
    gpo.po_header_field.restype = STRING
    gpo.po_header_set_field.argtypes = [STRING, STRING, STRING]
    gpo.po_header_set_field.restype = STRING

    # Locations (filepos)
    gpo.po_filepos_file.argtypes = [po_filepos_t]
    gpo.po_filepos_file.restype = STRING
    gpo.po_filepos_start_line.argtypes = [po_filepos_t]
    gpo.po_filepos_start_line.restype = c_int  # not strictly true casting
    gpo.po_message_filepos.argtypes = [po_message_t, c_int]
    gpo.po_message_filepos.restype = po_filepos_t
    gpo.po_message_add_filepos.argtypes = [po_message_t, STRING, c_size_t]
    gpo.po_message_remove_filepos.argtypes = [po_message_t, c_size_t]

    # Iterators
    gpo.po_message_iterator.argtypes = [po_file_t, STRING]
    gpo.po_message_iterator.restype = po_iterator_t
    gpo.po_message_iterator_free.argtypes = [po_iterator_t]
    gpo.po_next_message.argtypes = [po_iterator_t]
    gpo.po_next_message.restype = po_message_t
    gpo.po_message_insert.argtypes = [po_iterator_t, po_message_t]

    # Message (get methods)
    gpo.po_message_create.restype = po_message_t
    gpo.po_message_msgctxt.argtypes = [po_message_t]
    gpo.po_message_msgctxt.restype = STRING
    gpo.po_message_comments.argtypes = [po_message_t]
    gpo.po_message_comments.restype = STRING
    gpo.po_message_extracted_comments.argtypes = [po_message_t]
    gpo.po_message_extracted_comments.restype = STRING
    gpo.po_message_prev_msgctxt.argtypes = [po_message_t]
    gpo.po_message_prev_msgctxt.restype = STRING
    gpo.po_message_prev_msgid.argtypes = [po_message_t]
    gpo.po_message_prev_msgid.restype = STRING
    gpo.po_message_prev_msgid_plural.argtypes = [po_message_t]
    gpo.po_message_prev_msgid_plural.restype = STRING
    gpo.po_message_is_obsolete.argtypes = [po_message_t]
    gpo.po_message_is_obsolete.restype = c_int
    gpo.po_message_is_fuzzy.argtypes = [po_message_t]
    gpo.po_message_is_fuzzy.restype = c_int
    gpo.po_message_is_format.argtypes = [po_message_t, STRING]
    gpo.po_message_is_format.restype = c_int
    gpo.po_message_msgctxt.restype = STRING
    gpo.po_message_msgid.argtypes = [po_message_t]
    gpo.po_message_msgid.restype = STRING
    gpo.po_message_msgid_plural.argtypes = [po_message_t]
    gpo.po_message_msgid_plural.restype = STRING
    gpo.po_message_msgstr.argtypes = [po_message_t]
    gpo.po_message_msgstr.restype = STRING
    gpo.po_message_msgstr_plural.argtypes = [po_message_t, c_int]
    gpo.po_message_msgstr_plural.restype = STRING

    # Message (set methods)
    gpo.po_message_set_comments.argtypes = [po_message_t, STRING]
    gpo.po_message_set_extracted_comments.argtypes = [po_message_t, STRING]
    gpo.po_message_set_prev_msgctxt.argtypes = [po_message_t, STRING]
    gpo.po_message_set_prev_msgid.argtypes = [po_message_t, STRING]
    gpo.po_message_set_prev_msgid_plural.argtypes = [po_message_t, STRING]
    gpo.po_message_set_obsolete.argtypes = [po_message_t, c_int]
    gpo.po_message_set_fuzzy.argtypes = [po_message_t, c_int]
    gpo.po_message_set_format.argtypes = [po_message_t, STRING, c_int]
    gpo.po_message_set_msgctxt.argtypes = [po_message_t, STRING]
    gpo.po_message_set_msgid.argtypes = [po_message_t, STRING]
    gpo.po_message_set_msgstr.argtypes = [po_message_t, STRING]
    gpo.po_message_set_msgstr_plural.argtypes = [po_message_t, c_int, STRING]
    gpo.po_message_set_range.argtypes = [po_message_t, c_int, c_int]


# Load libgettextpo
gpo = None
# 'gettextpo' is recognised on Unix, while only 'libgettextpo' is recognised on
# windows. Therefore we test both.
names = ["gettextpo", "libgettextpo"]
for name in names:
    lib_location = ctypes.util.find_library(name)
    if lib_location:
        gpo = cdll.LoadLibrary(lib_location)
        if gpo:
            break
else:
    # Don't raise exception in Sphinx autodoc [where xml is Mock()ed]. There is
    # nothing special about use of xml here - any of the Mock classes set up
    # in docs/conf.py would work as well, but xml is likely always to be there.
    gpo = None
    if "xml" not in sys.modules or sys.modules["xml"].__path__ != "/dev/null":
        # Now we are getting desperate, so let's guess a unix type DLL that
        # might be in LD_LIBRARY_PATH or loaded with LD_PRELOAD
        try:
            gpo = cdll.LoadLibrary("libgettextpo.so")
        except OSError:
            raise ImportError("gettext PO library not found")

if gpo:
    setup_call_types(gpo)

# Setup the po_xerror_handler
xerror_handler = po_xerror_handler()
xerror_handler.xerror = xerror_prototype(xerror_cb)
xerror_handler.xerror2 = xerror2_prototype(xerror2_cb)


def escapeforpo(text):
    return pypo.escapeforpo(text)


def quoteforpo(text):
    return pypo.quoteforpo(text)


def unquotefrompo(postr):
    return pypo.unquotefrompo(postr)


def get_libgettextpo_version():
    """
    Returns the libgettextpo version.

    :rtype: three-value tuple
    :return: libgettextpo version in the following format::
        (major version, minor version, subminor version)
    """
    libversion = c_int.in_dll(gpo, "libgettextpo_version")
    major = libversion.value >> 16
    minor = (libversion.value >> 8) & 0xFF
    subminor = libversion.value - (major << 16) - (minor << 8)
    return major, minor, subminor


def gpo_encode(value):
    return value.encode("utf-8") if isinstance(value, str) else value


def gpo_decode(value):
    if isinstance(value, str):
        return value
    if isinstance(value, bytes):
        return value.decode("utf-8")
    return value


class pounit(pocommon.pounit):
    #: fixed encoding that is always used for cPO structure (self._gpo_message)
    CPO_ENC = "utf-8"

    def __init__(self, source=None, encoding="utf-8", gpo_message=None):
        self._rich_source = None
        self._rich_target = None
        encoding = encoding or "utf-8"
        if not gpo_message:
            self._gpo_message = gpo.po_message_create()
        if isinstance(source, str):
            self.source = source
            self.target = ""
        elif gpo_message:
            if encoding.lower() != self.CPO_ENC:
                features = ["msgctxt", "msgid", "msgid_plural"]
                features += ["prev_" + x for x in features]
                features += ["comments", "extracted_comments", "msgstr"]
                for feature in features:
                    text = getattr(gpo, "po_message_" + feature)(gpo_message)
                    if text:
                        getattr(gpo, "po_message_set_" + feature)(
                            gpo_message, text.decode(encoding).encode(self.CPO_ENC)
                        )
                # Also iterate through plural forms
                nplural = 0
                text = True
                while text:
                    text = gpo.po_message_msgstr_plural(gpo_message, nplural)
                    if text:
                        gpo.po_message_set_msgstr_plural(
                            gpo_message,
                            text.decode(encoding).encode(self.CPO_ENC),
                            nplural,
                        )
                    nplural += 1
            self._gpo_message = gpo_message
        self.infer_state()

    def infer_state(self):
        # FIXME: do obsolete
        if gpo.po_message_is_obsolete(self._gpo_message):
            if gpo.po_message_is_fuzzy(self._gpo_message):
                self.set_state_n(self.STATE[self.S_FUZZY_OBSOLETE][0])
            else:
                self.set_state_n(self.STATE[self.S_OBSOLETE][0])
        elif gpo.po_message_is_fuzzy(self._gpo_message):
            self.set_state_n(self.STATE[self.S_FUZZY][0])
        elif self.target:
            self.set_state_n(self.STATE[self.S_TRANSLATED][0])
        else:
            self.set_state_n(self.STATE[self.S_UNTRANSLATED][0])

    def setmsgid_plural(self, msgid_plural):
        if isinstance(msgid_plural, list):
            msgid_plural = "".join(msgid_plural)
        gpo.po_message_set_msgid_plural(self._gpo_message, gpo_encode(msgid_plural))

    msgid_plural = property(None, setmsgid_plural)

    @property
    def source(self):
        def remove_msgid_comments(text):
            if not text:
                return text
            if text.startswith("_:"):
                remainder = re.search(r"_: .*\n(.*)", text)
                if remainder:
                    return remainder.group(1)
                return ""
            return text

        singular = remove_msgid_comments(
            gpo_decode(gpo.po_message_msgid(self._gpo_message)) or ""
        )
        if singular:
            if self.hasplural():
                multi = multistring(singular)
                pluralform = (
                    gpo_decode(gpo.po_message_msgid_plural(self._gpo_message)) or ""
                )
                multi.extra_strings.append(pluralform)
                return multi
            return singular
        return ""

    @source.setter
    def source(self, source):
        if isinstance(source, multistring):
            source = source.strings
        if isinstance(source, list):
            gpo.po_message_set_msgid(self._gpo_message, gpo_encode(source[0]))
            if len(source) > 1:
                gpo.po_message_set_msgid_plural(
                    self._gpo_message, gpo_encode(source[1])
                )
        else:
            gpo.po_message_set_msgid(self._gpo_message, gpo_encode(source))
            gpo.po_message_set_msgid_plural(self._gpo_message, None)

    @property
    def target(self):
        if self.hasplural():
            plurals = []
            nplural = 0
            plural = gpo.po_message_msgstr_plural(self._gpo_message, nplural)
            while plural:
                plurals.append(plural.decode(self.CPO_ENC))
                nplural += 1
                plural = gpo.po_message_msgstr_plural(self._gpo_message, nplural)
            multi = multistring(plurals) if plurals else multistring("")
        else:
            multi = gpo_decode(gpo.po_message_msgstr(self._gpo_message)) or ""
        return multi

    @target.setter
    def target(self, target):
        # for plural strings: convert 'target' into a list
        if self.hasplural():
            if isinstance(target, multistring):
                target = target.strings
            elif isinstance(target, str):
                target = [target]
        # for non-plurals: check number of items in 'target'
        elif isinstance(target, (dict, list)):
            if len(target) == 1:
                target = target[0]
            else:
                raise ValueError(
                    "po msgid element has no plural but msgstr has %d elements (%s)"
                    % (len(target), target)
                )
        # empty the previous list of messages
        # TODO: the "pypo" implementation does not remove the previous items of
        #   the target, if self.target == target (essentially: comparing only
        #   the first item of a plural string with the single new string)
        #   Maybe this behaviour should be unified.
        if isinstance(target, (dict, list)):
            i = 0
            message = gpo.po_message_msgstr_plural(self._gpo_message, i)
            while message is not None:
                gpo.po_message_set_msgstr_plural(self._gpo_message, i, None)
                i += 1
                message = gpo.po_message_msgstr_plural(self._gpo_message, i)
        # add the items of a list
        if isinstance(target, list):
            for i, targetstring in enumerate(target):
                gpo.po_message_set_msgstr_plural(
                    self._gpo_message, i, gpo_encode(targetstring)
                )
        # add the values of a dict
        elif isinstance(target, dict):
            for i, targetstring in enumerate(target.values()):
                gpo.po_message_set_msgstr_plural(
                    self._gpo_message, i, gpo_encode(targetstring)
                )
        # add a single string
        elif target is None:
            gpo.po_message_set_msgstr(self._gpo_message, gpo_encode(""))
        else:
            gpo.po_message_set_msgstr(self._gpo_message, gpo_encode(target))

    def getid(self):
        """
        The unique identifier for this unit according to the conventions in
        .mo files.
        """
        id = gpo_decode(gpo.po_message_msgid(self._gpo_message)) or ""
        # Gettext does not consider the plural to determine duplicates, only
        # the msgid. For generation of .mo files, we might want to use this
        # code to generate the entry for the hash table, but for now, it is
        # commented out for conformance to gettext.
        #        plural = gpo.po_message_msgid_plural(self._gpo_message)
        #        if not plural is None:
        #            id = '%s\0%s' % (id, plural)
        context = gpo.po_message_msgctxt(self._gpo_message)
        if context:
            id = f"{gpo_decode(context)}\04{id}"
        return id

    def getnotes(self, origin=None):
        if origin is None:
            comments = gpo.po_message_comments(
                self._gpo_message
            ) + gpo.po_message_extracted_comments(self._gpo_message)
        elif origin == "translator":
            comments = gpo.po_message_comments(self._gpo_message)
        elif origin in {"programmer", "developer", "source code"}:
            comments = gpo.po_message_extracted_comments(self._gpo_message)
        else:
            raise ValueError("Comment type not valid")

        if comments and get_libgettextpo_version() < (0, 17, 0):
            comments = "\n".join(line for line in comments.split("\n"))
        # Let's drop the last newline
        return gpo_decode(comments[:-1])

    def addnote(self, text, origin=None, position="append"):
        # ignore empty strings and strings without non-space characters
        if not (text and text.strip()):
            return
        oldnotes = self.getnotes(origin)
        newnotes = None
        if oldnotes:
            if position == "append":
                newnotes = oldnotes + "\n" + text
            elif position == "merge":
                if oldnotes != text:
                    oldnoteslist = oldnotes.split("\n")
                    for newline in text.split("\n"):
                        newline = newline.rstrip("\r")
                        # avoid duplicate comment lines (this might cause some problems)
                        if newline not in oldnotes or len(newline) < 5:
                            oldnoteslist.append(newline)
                    newnotes = "\n".join(oldnoteslist)
            else:
                newnotes = text + "\n" + oldnotes
        else:
            newnotes = "\n".join(line.rstrip("\r") for line in text.split("\n"))

        if newnotes:
            newlines = []
            needs_space = get_libgettextpo_version() < (0, 17, 0)
            for line in newnotes.split("\n"):
                if line and needs_space:
                    newlines.append(" " + line)
                else:
                    newlines.append(line)
            newnotes = gpo_encode("\n".join(newlines))
            if origin in {"programmer", "developer", "source code"}:
                gpo.po_message_set_extracted_comments(self._gpo_message, newnotes)
            else:
                gpo.po_message_set_comments(self._gpo_message, newnotes)

    def removenotes(self, origin=None):
        gpo.po_message_set_comments(self._gpo_message, b"")

    def copy(self):
        newpo = self.__class__()
        newpo._gpo_message = self._gpo_message
        return newpo

    def merge(self, otherpo, overwrite=False, comments=True, authoritative=False):
        """
        Merges the otherpo (with the same msgid) into this one.

        Overwrite non-blank self.msgstr only if overwrite is True
        merge comments only if comments is True
        """
        if not isinstance(otherpo, pounit):
            super().merge(otherpo, overwrite, comments)
            return
        if comments:
            self.addnote(
                otherpo.getnotes("translator"), origin="translator", position="merge"
            )
            # FIXME mergelists(self.typecomments, otherpo.typecomments)
            if not authoritative:
                # We don't bring across otherpo.automaticcomments as we consider ourself
                # to be the the authority.  Same applies to otherpo.msgidcomments
                self.addnote(
                    otherpo.getnotes("developer"), origin="developer", position="merge"
                )
                self.msgidcomment = otherpo._extract_msgidcomments() or None
                self.addlocations(otherpo.getlocations())
        if not self.istranslated() or overwrite:
            # Remove kde-style comments from the translation (if any).
            if self._extract_msgidcomments(otherpo.target):
                otherpo.target = otherpo.target.replace(
                    "_: " + otherpo._extract_msgidcomments() + "\n", ""
                )
            self.target = otherpo.target
            if (
                self.source != otherpo.source
                or self.getcontext() != otherpo.getcontext()
            ):
                self.markfuzzy()
            else:
                self.markfuzzy(otherpo.isfuzzy())
        elif not otherpo.istranslated():
            if self.source != otherpo.source:
                self.markfuzzy()
        elif self.target != otherpo.target:
            self.markfuzzy()

    def isheader(self):
        # return self.source == "" and self.target != ""
        # we really want to make sure that there is no msgidcomment or msgctxt
        return not self.getid() and len(self.target) > 0

    def isblank(self):
        return len(self.source) == len(self.target) == len(self.getcontext()) == 0

    def hastypecomment(self, typecomment):
        return gpo.po_message_is_format(self._gpo_message, gpo_encode(typecomment))

    def settypecomment(self, typecomment, present=True):
        gpo.po_message_set_format(self._gpo_message, gpo_encode(typecomment), present)

    def hasmarkedcomment(self, commentmarker):
        commentmarker = f"({commentmarker})"
        for comment in self.getnotes("translator").split("\n"):
            if comment.startswith(commentmarker):
                return True
        return False

    def isfuzzy(self):
        return gpo.po_message_is_fuzzy(self._gpo_message)

    def _domarkfuzzy(self, present=True):
        gpo.po_message_set_fuzzy(self._gpo_message, present)

    def makeobsolete(self):
        # FIXME: libgettexpo currently does not reset other data, we probably want to do that
        # but a better solution would be for libgettextpo to output correct data on serialisation
        gpo.po_message_set_obsolete(self._gpo_message, True)
        self.infer_state()

    def resurrect(self):
        gpo.po_message_set_obsolete(self._gpo_message, False)
        self.infer_state()

    def hasplural(self):
        return gpo.po_message_msgid_plural(self._gpo_message) is not None

    def _extract_msgidcomments(self, text=None):
        """
        Extract KDE style msgid comments from the unit.

        :rtype: String
        :return: Returns the extracted msgidcomments found in this unit's msgid.
        """
        if not text:
            text = gpo_decode(gpo.po_message_msgid(self._gpo_message)) or ""
        if text:
            return pocommon.extract_msgid_comment(text)
        return ""

    def setmsgidcomment(self, msgidcomment):
        if msgidcomment:
            self.source = f"_: {msgidcomment}\n{self.source}"

    msgidcomment = property(_extract_msgidcomments, setmsgidcomment)

    def __str__(self):
        pf = pofile(noheader=True)
        pf.addunit(self)
        return bytes(pf).decode(self.CPO_ENC)

    def getlocations(self):
        locations = []
        i = 0
        location = gpo.po_message_filepos(self._gpo_message, i)
        while location:
            locname = gpo_decode(gpo.po_filepos_file(location))
            locline = gpo.po_filepos_start_line(location)
            locstring = locname if locline == -1 else ":".join([locname, str(locline)])
            locations.append(pocommon.unquote_plus(locstring))
            i += 1
            location = gpo.po_message_filepos(self._gpo_message, i)
        return locations

    def addlocation(self, location):
        if location.find(" ") != -1:
            location = pocommon.quote_plus(location)
        parts = location.split(":")
        if len(parts) == 2 and parts[1].isdigit():
            file = parts[0]
            line = int(parts[1] or "0")
        else:
            file = location
            line = -1
        gpo.po_message_add_filepos(self._gpo_message, gpo_encode(file), line)

    def getcontext(self):
        msgctxt = gpo.po_message_msgctxt(self._gpo_message)
        if msgctxt:
            return gpo_decode(msgctxt)
        return self._extract_msgidcomments()

    def setcontext(self, context):
        gpo.po_message_set_msgctxt(self._gpo_message, gpo_encode(context))

    @classmethod
    def buildfromunit(cls, unit, encoding=None):
        """
        Build a native unit from a foreign unit, preserving as much
        information as possible.
        """
        if type(unit) is cls and hasattr(unit, "copy") and callable(unit.copy):
            return unit.copy()
        if isinstance(unit, pocommon.pounit):
            newunit = cls(unit.source, encoding)
            newunit.target = unit.target
            # context
            newunit.msgidcomment = unit._extract_msgidcomments()
            context = unit.getcontext()
            if not newunit.msgidcomment and context:
                newunit.setcontext(context)

            locations = unit.getlocations()
            if locations:
                newunit.addlocations(locations)
            notes = unit.getnotes("developer")
            if notes:
                newunit.addnote(notes, "developer")
            notes = unit.getnotes("translator")
            if notes:
                newunit.addnote(notes, "translator")
            if unit.isobsolete():
                newunit.makeobsolete()
            newunit.markfuzzy(unit.isfuzzy())
            for tc in ["python-format", "c-format", "php-format"]:
                if unit.hastypecomment(tc):
                    newunit.settypecomment(tc)
                    # We assume/guess/hope that there will only be one
                    break
            return newunit
        return base.TranslationUnit.buildfromunit(unit)


class pofile(pocommon.pofile):
    UnitClass = pounit

    def __init__(self, inputfile=None, noheader=False, **kwargs):
        self._gpo_memory_file = None
        self._gpo_message_iterator = None
        self.sourcelanguage = None
        self.targetlanguage = None
        if inputfile is None:
            self.units = []
            self._encoding = kwargs.get("encoding")
            self._gpo_memory_file = gpo.po_file_create()
            self._gpo_message_iterator = gpo.po_message_iterator(
                self._gpo_memory_file, None
            )
        super().__init__(inputfile=inputfile, noheader=noheader, **kwargs)

    def addunit(self, unit, new=True):
        if new:
            gpo.po_message_insert(self._gpo_message_iterator, unit._gpo_message)
        super().addunit(unit)

    def removeunit(self, unit):
        # There seems to be no API to remove a message
        raise ValueError("Unit removal not supported by cpo")

    def _insert_header(self, header):
        header._store = self
        self.units.insert(0, header)
        self._free_iterator()
        self._gpo_message_iterator = gpo.po_message_iterator(
            self._gpo_memory_file, None
        )
        gpo.po_message_insert(self._gpo_message_iterator, header._gpo_message)
        while gpo.po_next_message(self._gpo_message_iterator):
            pass

    def removeduplicates(self, duplicatestyle="merge"):
        """Make sure each msgid is unique ; merge comments etc from duplicates into original."""
        # TODO: can we handle consecutive calls to removeduplicates()? What
        # about files already containing msgctxt? - test
        id_dict = {}
        uniqueunits = []
        # TODO: this is using a list as the pos aren't hashable, but this is slow.
        # probably not used frequently enough to worry about it, though.
        markedpos = []

        def addcomment(thepo):
            thepo.msgidcomment = " ".join(thepo.getlocations())
            markedpos.append(thepo)

        for thepo in self.units:
            id = thepo.getid()
            if thepo.isheader() and not thepo.getlocations():
                # header msgids shouldn't be merged...
                uniqueunits.append(thepo)
            elif id in id_dict:
                if duplicatestyle == "merge":
                    if id:
                        id_dict[id].merge(thepo)
                    else:
                        addcomment(thepo)
                        uniqueunits.append(thepo)
                elif duplicatestyle == "msgctxt":
                    origpo = id_dict[id]
                    if origpo not in markedpos:
                        origpo.setcontext(" ".join(origpo.getlocations()))
                        markedpos.append(thepo)
                    thepo.setcontext(" ".join(thepo.getlocations()))
                    thepo_msgctxt = gpo.po_message_msgctxt(thepo._gpo_message)
                    idpo_msgctxt = gpo.po_message_msgctxt(id_dict[id]._gpo_message)
                    if thepo_msgctxt != idpo_msgctxt:
                        uniqueunits.append(thepo)
                    else:
                        logger.warning(
                            "Duplicate unit found with msgctx of '%s' and source '%s'",
                            thepo_msgctxt,
                            thepo.source,
                        )
            else:
                if not id:
                    if duplicatestyle == "merge":
                        addcomment(thepo)
                    else:
                        thepo.setcontext(" ".join(thepo.getlocations()))
                id_dict[id] = thepo
                uniqueunits.append(thepo)
        new_gpo_memory_file = gpo.po_file_create()
        new_gpo_message_iterator = gpo.po_message_iterator(new_gpo_memory_file, None)
        for unit in uniqueunits:
            gpo.po_message_insert(new_gpo_message_iterator, unit._gpo_message)
        self._free_iterator()
        self._gpo_message_iterator = new_gpo_message_iterator
        self._free_memory_file()
        self._gpo_memory_file = new_gpo_memory_file
        self.units = uniqueunits

    def serialize(self, out):
        def obsolete_workaround():
            # Remove all items that are not output by msgmerge when a unit is obsolete.  This is a work
            # around for bug in libgettextpo
            # FIXME Do version test in case they fix this bug
            for unit in self.units:
                if unit.isobsolete():
                    gpo.po_message_set_extracted_comments(unit._gpo_message, b"")
                    location = gpo.po_message_filepos(unit._gpo_message, 0)
                    while location:
                        gpo.po_message_remove_filepos(unit._gpo_message, 0)
                        location = gpo.po_message_filepos(unit._gpo_message, 0)

        def writefile(filename):
            xerror_storage.exception = None
            result = gpo.po_file_write_v2(
                self._gpo_memory_file, gpo_encode(filename), xerror_handler
            )
            if xerror_storage.exception is not None:
                raise xerror_storage.exception
            if result is None:
                raise ValueError("Unknown error while saving file")
            with open(filename, "rb") as tfile:
                return tfile.read()

        outputstring = ""
        if self._gpo_memory_file:
            obsolete_workaround()
            f, fname = tempfile.mkstemp(prefix="translate", suffix=".po")
            os.close(f)
            try:
                outputstring = writefile(fname)
                if self.encoding != pounit.CPO_ENC:
                    try:
                        outputstring = outputstring.decode(pounit.CPO_ENC).encode(
                            self.encoding
                        )
                    except UnicodeEncodeError:
                        self.encoding = pounit.CPO_ENC
                        self.updateheader(
                            content_type="text/plain; charset=UTF-8",
                            content_transfer_encoding="8bit",
                        )
                        outputstring = writefile(fname)
            finally:
                os.remove(fname)
        out.write(outputstring)

    def isempty(self):
        """Returns True if the object doesn't contain any translation units."""
        if len(self.units) == 0:
            return True
        # Skip the first unit if it is a header.
        units = self.units[1:] if self.units[0].isheader() else self.units

        return all(not (not unit.isblank() and not unit.isobsolete()) for unit in units)

    def parse(self, input):
        if hasattr(input, "name"):
            self.filename = input.name
        elif not getattr(self, "filename", ""):
            self.filename = ""

        if hasattr(input, "read"):
            posrc = input.read()
            input.close()
            input = posrc

        needtmpfile = not os.path.isfile(input)
        if needtmpfile:
            # This is not a file - we write the string to a temporary file
            fd, fname = tempfile.mkstemp(prefix="translate", suffix=".po")
            os.write(fd, input)
            input = fname
            os.close(fd)

        try:
            xerror_storage.exception = None
            self._free_memory_file()
            self._gpo_memory_file = gpo.po_file_read_v3(
                gpo_encode(input), xerror_handler
            )
            if xerror_storage.exception is not None:
                raise xerror_storage.exception
            if self._gpo_memory_file is None:
                logger.error("Error:")
        finally:
            if needtmpfile:
                os.remove(input)

        self.units = []
        # Handle xerrors here
        self._header = gpo.po_file_domain_header(self._gpo_memory_file, None)
        if self._header:
            charset = gpo_decode(
                gpo.po_header_field(self._header, gpo_encode("Content-Type"))
            )
            if charset:
                charset = re.search(r"charset=([^\s]+)", charset).group(1)
            self.encoding = charset
        self._free_iterator()
        self._gpo_message_iterator = gpo.po_message_iterator(
            self._gpo_memory_file, None
        )
        newmessage = gpo.po_next_message(self._gpo_message_iterator)
        while newmessage:
            newunit = pounit(gpo_message=newmessage, encoding=self.encoding)
            self.addunit(newunit, new=False)
            newmessage = gpo.po_next_message(self._gpo_message_iterator)
        self._free_iterator()

    def __del__(self):
        self._free_iterator()
        self._free_memory_file()

    def _free_memory_file(self):
        return
        if self._gpo_memory_file is not None:
            gpo.po_file_free(self._gpo_memory_file)
            self._gpo_memory_file = None

    def _free_iterator(self):
        if self._gpo_message_iterator is not None:
            gpo.po_message_iterator_free(self._gpo_message_iterator)
            self._gpo_message_iterator = None



from translate.storage.placeables import general


def test_placeable_numbers():
    """Check the correct functioning of number placeables."""
    assert general.NumberPlaceable(["25"]) in general.NumberPlaceable.parse(
        "Here is a 25 number"
    )
    assert general.NumberPlaceable(["-25"]) in general.NumberPlaceable.parse(
        "Here is a -25 number"
    )
    assert general.NumberPlaceable(["+25"]) in general.NumberPlaceable.parse(
        "Here is a +25 number"
    )
    assert general.NumberPlaceable(["25.00"]) in general.NumberPlaceable.parse(
        "Here is a 25.00 number"
    )
    assert general.NumberPlaceable(["2,500.00"]) in general.NumberPlaceable.parse(
        "Here is a 2,500.00 number"
    )
    assert general.NumberPlaceable(["1\u00a0000,99"]) in general.NumberPlaceable.parse(
        "Here is a 1\u00a0000,99 number"
    )


def test_placeable_newline():
    assert general.NewlinePlaceable.parse("A newline\n")[1] == general.NewlinePlaceable(
        ["\n"]
    )
    assert general.NewlinePlaceable.parse("First\nSecond")[
        1
    ] == general.NewlinePlaceable(["\n"])


def test_placeable_alt_attr():
    assert general.AltAttrPlaceable.parse(
        'Click on the <img src="image.jpg" alt="Image">'
    )[1] == general.AltAttrPlaceable(['alt="Image"'])


def test_placeable_qt_formatting():
    assert general.QtFormattingPlaceable.parse("One %1 %99 %L1 are all valid")[
        1
    ] == general.QtFormattingPlaceable(["%1"])
    assert general.QtFormattingPlaceable.parse("One %1 %99 %L1 are all valid")[
        3
    ] == general.QtFormattingPlaceable(["%99"])
    assert general.QtFormattingPlaceable.parse("One %1 %99 %L1 are all valid")[
        5
    ] == general.QtFormattingPlaceable(["%L1"])


def test_placeable_camelcase():
    assert general.CamelCasePlaceable.parse("CamelCase")[
        0
    ] == general.CamelCasePlaceable(["CamelCase"])
    assert general.CamelCasePlaceable.parse("iPod")[0] == general.CamelCasePlaceable(
        ["iPod"]
    )
    assert general.CamelCasePlaceable.parse("DokuWiki")[
        0
    ] == general.CamelCasePlaceable(["DokuWiki"])
    assert general.CamelCasePlaceable.parse("KBabel")[0] == general.CamelCasePlaceable(
        ["KBabel"]
    )
    assert general.CamelCasePlaceable.parse("_Bug") is None
    assert general.CamelCasePlaceable.parse("NOTCAMEL") is None


def test_placeable_space():
    assert general.SpacesPlaceable.parse(" Space at start")[
        0
    ] == general.SpacesPlaceable([" "])
    assert general.SpacesPlaceable.parse("Space at end ")[1] == general.SpacesPlaceable(
        [" "]
    )
    assert general.SpacesPlaceable.parse("Double  space")[1] == general.SpacesPlaceable(
        ["  "]
    )


def test_placeable_punctuation():
    assert (
        general.PunctuationPlaceable.parse(
            'These, are not. Special: punctuation; marks! Or are "they"?'
        )
        is None
    )
    assert general.PunctuationPlaceable.parse("Downloading…")[
        1
    ] == general.PunctuationPlaceable(["…"])


def test_placeable_xml_entity():
    assert general.XMLEntityPlaceable.parse("&brandShortName;")[
        0
    ] == general.XMLEntityPlaceable(["&brandShortName;"])
    assert general.XMLEntityPlaceable.parse("&#1234;")[0] == general.XMLEntityPlaceable(
        ["&#1234;"]
    )
    assert general.XMLEntityPlaceable.parse("&xDEAD;")[0] == general.XMLEntityPlaceable(
        ["&xDEAD;"]
    )


def test_placeable_xml_tag():
    assert general.XMLTagPlaceable.parse("<a>koei</a>")[0] == general.XMLTagPlaceable(
        ["<a>"]
    )
    assert general.XMLTagPlaceable.parse("<a>koei</a>")[2] == general.XMLTagPlaceable(
        ["</a>"]
    )
    assert general.XMLTagPlaceable.parse("<Exif.XResolution>")[
        0
    ] == general.XMLTagPlaceable(["<Exif.XResolution>"])
    assert general.XMLTagPlaceable.parse("<tag_a>")[0] == general.XMLTagPlaceable(
        ["<tag_a>"]
    )
    assert general.XMLTagPlaceable.parse('<img src="koei.jpg" />')[
        0
    ] == general.XMLTagPlaceable(['<img src="koei.jpg" />'])
    # We don't want this to be recognised, so we test for None - not sure if that is a stable assumption
    assert general.XMLTagPlaceable.parse("<important word>") is None
    assert general.XMLTagPlaceable.parse('<img ="koei.jpg" />') is None
    assert general.XMLTagPlaceable.parse('<img "koei.jpg" />') is None
    assert general.XMLTagPlaceable.parse('<span xml:space="preserve">')[
        0
    ] == general.XMLTagPlaceable(['<span xml:space="preserve">'])
    assert general.XMLTagPlaceable.parse(
        '<img src="http://translate.org.za/blogs/friedel/sites/translate.org.za.blogs.friedel/files/virtaal-7f_help.png" alt="Virtaal met lêernaam-pseudovertaling" style="border: 1px dotted grey;" />'
    )[0] == general.XMLTagPlaceable(
        [
            '<img src="http://translate.org.za/blogs/friedel/sites/translate.org.za.blogs.friedel/files/virtaal-7f_help.png" alt="Virtaal met lêernaam-pseudovertaling" style="border: 1px dotted grey;" />'
        ]
    )
    # Bug 933
    assert general.XMLTagPlaceable.parse(
        'This entry expires in %days% days. Would you like to <a href="%href%?PHPSESSID=5d59c559cf4eb9f1d278918271fbe68a" title="Renew this Entry Now">Renew this Entry Now</a> ?'
    )[1] == general.XMLTagPlaceable(
        [
            '<a href="%href%?PHPSESSID=5d59c559cf4eb9f1d278918271fbe68a" title="Renew this Entry Now">'
        ]
    )
    assert general.XMLTagPlaceable.parse(
        """<span weight='bold' size='larger'>Your Google Account is locked</span>"""
    )[0] == general.XMLTagPlaceable(["""<span weight='bold' size='larger'>"""])


def test_placeable_option():
    assert general.OptionPlaceable.parse("Type --help for this help")[
        1
    ] == general.OptionPlaceable(["--help"])
    assert general.OptionPlaceable.parse("Short -S ones also")[
        1
    ] == general.OptionPlaceable(["-S"])


def test_placeable_file():
    assert general.FilePlaceable.parse("Store in /home/user")[
        1
    ] == general.FilePlaceable(["/home/user"])
    assert general.FilePlaceable.parse("Store in ~/Download directory")[
        1
    ] == general.FilePlaceable(["~/Download"])


def test_placeable_email():
    assert general.EmailPlaceable.parse("Send email to info@example.com")[
        1
    ] == general.EmailPlaceable(["info@example.com"])
    assert general.EmailPlaceable.parse("Send email to mailto:info@example.com")[
        1
    ] == general.EmailPlaceable(["mailto:info@example.com"])


def test_placeable_caps():
    assert general.CapsPlaceable.parse("Use the HTML page")[1] == general.CapsPlaceable(
        ["HTML"]
    )
    assert general.CapsPlaceable.parse("I am") is None
    assert general.CapsPlaceable.parse("Use the A4 paper") is None
    assert general.CapsPlaceable.parse("In GTK+")[1] == general.CapsPlaceable(["GTK+"])
    #    assert general.CapsPlaceable.parse('GNOME-stuff')[0] == general.CapsPlaceable(['GNOME'])
    assert general.CapsPlaceable.parse("with XDG_USER_DIRS")[
        1
    ] == general.CapsPlaceable(["XDG_USER_DIRS"])


def test_placeable_formatting():
    fp = general.FormattingPlaceable
    assert fp.parse("There were %d cows")[1] == fp(["%d"])
    assert fp.parse("There were %Id cows")[1] == fp(["%Id"])
    assert fp.parse("There were %d %s")[3] == fp(["%s"])
    assert fp.parse("%1$s was kicked by %2$s")[0] == fp(["%1$s"])
    assert fp.parse("There were %Id cows")[1] == fp(["%Id"])
    assert fp.parse("There were % d cows")[1] == fp(["% d"])
    # only a real space is allowed as formatting flag
    assert fp.parse("There were %\u00a0d cows") is None
    assert fp.parse("There were %'f cows")[1] == fp(["%'f"])
    assert fp.parse("There were %#x cows")[1] == fp(["%#x"])

    # field width
    assert fp.parse("There were %3d cows")[1] == fp(["%3d"])
    assert fp.parse("There were %33d cows")[1] == fp(["%33d"])
    assert fp.parse("There were %*d cows")[1] == fp(["%*d"])

    # numbered variables
    assert fp.parse("There were %1$d cows")[1] == fp(["%1$d"])


def test_placeable_doubleat():
    dap = general.DoubleAtPlaceable
    assert dap.parse("There were @@number@@ cows")[1] == dap(["@@number@@"])
    assert dap.parse("There were @@number1@@ cows and @@number2@@ sheep")[1] == dap(
        ["@@number1@@"]
    )
    assert dap.parse("There were @@number1@@ cows and @@number2@@ sheep")[3] == dap(
        ["@@number2@@"]
    )


def test_placeable_brace():
    bp = general.BracePlaceable
    # Double braces
    assert bp.parse("There were {{number}} cows")[1] == bp(["{{number}}"])
    assert bp.parse("There were {{number1}} cows and {{number2}} sheep")[1] == bp(
        ["{{number1}}"]
    )
    assert bp.parse("There were {{number1}} cows and {{number2}} sheep")[3] == bp(
        ["{{number2}}"]
    )

    # Single braces
    assert bp.parse("There were {number} cows")[1] == bp(["{number}"])
    assert bp.parse("There were {number1} cows and {number2} sheep")[1] == bp(
        ["{number1}"]
    )
    assert bp.parse("There were {number1} cows and {number2} sheep")[3] == bp(
        ["{number2}"]
    )

    # Mixed single and double braces
    assert bp.parse("There were {number1} cows and {{number2}} sheep")[1] == bp(
        ["{number1}"]
    )
    assert bp.parse("There were {number1} cows and {{number2}} sheep")[3] == bp(
        ["{{number2}}"]
    )
    assert bp.parse("There were {{number1}} cows and {number2} sheep")[1] == bp(
        ["{{number1}}"]
    )
    assert bp.parse("There were {{number1}} cows and {number2} sheep")[3] == bp(
        ["{number2}"]
    )


def test_python_placeable():
    pfp = general.PythonFormattingPlaceable
    # No conversion
    assert pfp.parse("100%% correct")[1] == pfp(["%%"])

    # Mapping keys
    assert pfp.parse("There were %(number)d cows")[1] == pfp(["%(number)d"])
    assert pfp.parse("There were %(cows.number)d cows")[1] == pfp(["%(cows.number)d"])
    assert pfp.parse("There were %(number of cows)d cows")[1] == pfp(
        ["%(number of cows)d"]
    )

    # Conversion flags
    assert pfp.parse("There were %(number)03d cows")[1] == pfp(["%(number)03d"])
    assert pfp.parse("There were %(number) 3d cows")[1] == pfp(["%(number) 3d"])

    # Minimum field width
    assert pfp.parse("There were %(number)*d cows")[1] == pfp(["%(number)*d"])

    # Precision
    assert pfp.parse("There were %(number)3.1d cows")[1] == pfp(["%(number)3.1d"])

    # Length modifier
    assert pfp.parse("There were %(number)Ld cows")[1] == pfp(["%(number)Ld"])


# TODO: JavaMessageFormatPlaceable, UrlPlaceable, XMLTagPlaceable



#
# Copyright 2002-2007 Zuza Software Foundation
# Copyright 2016 F Wolff
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.


"""
From the GNU gettext manual:
     WHITE-SPACE
     #  TRANSLATOR-COMMENTS
     #. AUTOMATIC-COMMENTS
     #| PREVIOUS MSGID                 (Gettext 0.16 - check if this is the correct position - not yet implemented)
     #: REFERENCE...
     #, FLAG...
     #= FLAG...
     msgctxt CONTEXT                   (Gettext 0.15)
     msgid UNTRANSLATED-STRING
     msgstr TRANSLATED-STRING.
"""

import re

SINGLE_BYTE_ENCODING = "iso-8859-1"
isspace = str.isspace
find = str.find
rfind = str.rfind
startswith = str.startswith
append = list.append
decode = bytes.decode


class PoParseError(ValueError):
    def __init__(self, parse_state, message=None):
        self.parse_state = parse_state
        if message is None:
            message = "Syntax error"
        super().__init__(
            f"{message} on line {parse_state.lineno}: {parse_state.last_line!r}"
        )


class ParseState:
    def __init__(self, input_iterator, UnitClass, encoding=SINGLE_BYTE_ENCODING):
        # A single-byte encoding is first defined to be able to read the header
        # without risking UnicodeDecodeErrors. As soon as the header is parsed,
        # the encoding defined in the header is used for re-encoding the header
        # and for decoding all further strings.
        self._input_iterator = input_iterator
        self.next_line = ""
        self.last_line = ""
        self.lineno = 0
        self.eof = False
        self.encoding = encoding
        self.read_line()
        self.UnitClass = UnitClass

    def decode(self, string):
        if self.encoding is not None:
            return decode(string, self.encoding)
        return string

    def read_line(self):
        self.last_line = current = self.next_line
        if self.eof:
            return current
        try:
            self.next_line = next(self._input_iterator)
            self.lineno += 1
            while not self.eof and self.next_line.isspace():
                self.next_line = next(self._input_iterator)
                self.lineno += 1
        except StopIteration:
            self.next_line = ""
            self.eof = True
        else:
            if isinstance(self.next_line, bytes) and self.encoding is not None:
                self.next_line = decode(self.next_line, self.encoding)
        return current

    def new_input(self, input_):
        return ParseState(input_, self.UnitClass, self.encoding)


def read_prevmsgid_lines(parse_state):
    """
    Read all the lines belonging starting with #|. These lines contain the
    previous msgid and msgctxt info. We strip away the leading '#| ' and read
    until we stop seeing #|.
    """
    prevmsgid_lines = []
    next_line = parse_state.next_line
    while startswith(next_line, "#|") or startswith(next_line, "|"):
        content = parse_state.read_line()
        prefix_len = content.index("|") + 1
        while content[prefix_len] == " ":
            prefix_len += 1
        content = content[prefix_len:]
        append(prevmsgid_lines, content)
        next_line = parse_state.next_line
    return prevmsgid_lines


def parse_prev_msgctxt(parse_state, unit):
    parse_message(parse_state, "msgctxt", 7, unit.prev_msgctxt)
    return len(unit.prev_msgctxt) > 0


def parse_prev_msgid(parse_state, unit):
    parse_message(parse_state, "msgid", 5, unit.prev_msgid)
    return len(unit.prev_msgid) > 0


def parse_prev_msgid_plural(parse_state, unit):
    parse_message(parse_state, "msgid_plural", 12, unit.prev_msgid_plural)
    return len(unit.prev_msgid_plural) > 0


def parse_comment(parse_state, unit):
    next_line = parse_state.next_line.lstrip()
    if next_line and next_line[0] in {"#", "|"}:
        next_char = next_line[1]
        if next_char == ".":
            append(unit.automaticcomments, next_line)
        elif next_line[0] == "|" or next_char == "|":
            parsed = False
            # Read all the lines starting with #|
            prevmsgid_lines = read_prevmsgid_lines(parse_state)
            # Create a parse state object that holds these lines
            ps = parse_state.new_input(iter(prevmsgid_lines))
            # Parse the msgctxt if any
            parsed |= parse_prev_msgctxt(ps, unit)
            # Parse the msgid if any
            parsed |= parse_prev_msgid(ps, unit)
            # Parse the msgid_plural if any
            parsed |= parse_prev_msgid_plural(ps, unit)
            # Fail with error in csae nothing was parsed
            if not parsed:
                raise PoParseError(parse_state)
            return parse_state.next_line
        elif next_char == ":":
            append(unit.sourcecomments, next_line)
        elif next_char in {",", "="}:
            # One of these will be workflow flags and one sticky flags in the future,
            # normalize to #, for now
            append(unit.typecomments, f"#,{next_line[2:]}")
        elif next_char == "~":
            # Special case: we refuse to parse obsoletes: they are done
            # elsewhere to ensure we reuse the normal unit parsing code
            return None
        else:
            append(unit.othercomments, next_line)
        return parse_state.read_line()
    return None


def parse_comments(parse_state, unit):
    if not parse_comment(parse_state, unit):
        return None
    while parse_comment(parse_state, unit):
        pass
    return True


def read_obsolete_lines(parse_state):
    """Read all the lines belonging to the current unit if obsolete."""
    obsolete_lines = []
    next_line = parse_state.next_line
    while startswith(next_line, "#~"):
        content = parse_state.read_line()[2:].lstrip()
        append(obsolete_lines, content)
        next_line = parse_state.next_line
        if startswith(content, "msgstr"):
            # now we saw a msgstr, so we need to become more conservative to
            # avoid parsing into the following unit
            while startswith(next_line, '#~ "') or startswith(next_line, "#~ msgstr"):
                content = parse_state.read_line()[3:]
                append(obsolete_lines, content)
                next_line = parse_state.next_line
            break
    return obsolete_lines


def parse_obsolete(parse_state, unit):
    obsolete_lines = read_obsolete_lines(parse_state)
    if len(obsolete_lines) == 0:
        return None
    unit = parse_unit(parse_state.new_input(iter(obsolete_lines)), unit)
    if unit is not None:
        unit.makeobsolete()
    return unit


def parse_quoted(parse_state, start_pos=0):
    line = parse_state.next_line
    left = find(line, '"', start_pos)
    if left == start_pos or isspace(line[start_pos:left]):
        right = rfind(line, '"')
        if left != right:
            return parse_state.read_line()[left : right + 1]
        raise PoParseError(parse_state, "end-of-line within string")
    return None


def parse_msg_comment(parse_state, msg_comment_list, string):
    while string is not None:
        append(msg_comment_list, string)
        if find(string, "\\n") > -1:
            return parse_quoted(parse_state)
        string = parse_quoted(parse_state)
    return None


def parse_multiple_quoted(parse_state, msg_list, msg_comment_list, first_start_pos=0):
    string = parse_quoted(parse_state, first_start_pos)
    while string is not None:
        if msg_comment_list is None or not startswith(string, '"_:'):
            append(msg_list, string)
            string = parse_quoted(parse_state)
        else:
            string = parse_msg_comment(parse_state, msg_comment_list, string)


def parse_message(
    parse_state, start_of_string, start_of_string_len, msg_list, msg_comment_list=None
):
    if startswith(parse_state.next_line, start_of_string):
        return parse_multiple_quoted(
            parse_state, msg_list, msg_comment_list, start_of_string_len
        )
    return None


def parse_msgctxt(parse_state, unit):
    parse_message(parse_state, "msgctxt", 7, unit.msgctxt)
    return len(unit.msgctxt) > 0


def parse_msgid(parse_state, unit):
    parse_message(parse_state, "msgid", 5, unit.msgid, unit.msgidcomments)
    return len(unit.msgid) > 0 or len(unit.msgidcomments) > 0


def parse_msgstr(parse_state, unit):
    parse_message(parse_state, "msgstr", 6, unit.msgstr)
    return len(unit.msgstr) > 0


def parse_msgid_plural(parse_state, unit):
    parse_message(
        parse_state, "msgid_plural", 12, unit.msgid_plural, unit.msgid_pluralcomments
    )
    return len(unit.msgid_plural) > 0 or len(unit.msgid_pluralcomments) > 0


MSGSTR_ARRAY_ENTRY_LEN = len("msgstr[")


def add_to_dict(msgstr_dict, line, right_bracket_pos, entry):
    index = int(line[MSGSTR_ARRAY_ENTRY_LEN:right_bracket_pos])
    if index not in msgstr_dict:
        msgstr_dict[index] = []
    msgstr_dict[index].extend(entry)


def get_entry(parse_state, right_bracket_pos):
    entry = []
    parse_message(parse_state, "msgstr[", right_bracket_pos + 1, entry)
    return entry


def parse_msgstr_array_entry(parse_state, msgstr_dict):
    line = parse_state.next_line
    right_bracket_pos = find(line, "]", MSGSTR_ARRAY_ENTRY_LEN)
    if right_bracket_pos >= 0:
        entry = get_entry(parse_state, right_bracket_pos)
        if entry:
            add_to_dict(msgstr_dict, line, right_bracket_pos, entry)
            return True
        return False
    return False


def parse_msgstr_array(parse_state, unit):
    msgstr_dict = {}
    result = parse_msgstr_array_entry(parse_state, msgstr_dict)
    if not result:  # We require at least one result
        return False
    while parse_msgstr_array_entry(parse_state, msgstr_dict):
        pass
    unit.msgstr = msgstr_dict
    return True


def parse_plural(parse_state, unit):
    return bool(
        parse_msgid_plural(parse_state, unit) and parse_msgstr_array(parse_state, unit)
    )


def parse_msg_entries(parse_state, unit):
    parse_msgctxt(parse_state, unit)
    return bool(
        parse_msgid(parse_state, unit)
        and (parse_msgstr(parse_state, unit) or parse_plural(parse_state, unit))
    )


def parse_unit(parse_state, unit=None):
    unit = unit or parse_state.UnitClass()
    parsed_comments = parse_comments(parse_state, unit)
    obsolete_unit = parse_obsolete(parse_state, unit)
    if obsolete_unit is not None:
        return obsolete_unit
    parsed_msg_entries = parse_msg_entries(parse_state, unit)
    if parsed_comments or parsed_msg_entries:
        return unit
    return None


def set_encoding(parse_state, store, unit):
    charset = None
    if (
        isinstance(unit.msgstr, list)
        and unit.msgstr
        and isinstance(unit.msgstr[0], str)
    ):
        charset = re.search(r"charset=([^\s\\n]+)", "".join(unit.msgstr))
    if charset:
        encoding = charset.group(1)
        if encoding != "CHARSET":
            store._encoding = encoding
        else:
            store._encoding = "utf-8"
    else:
        store._encoding = "utf-8"
    parse_state.encoding = store._encoding


def decode_list(lst, decode):
    return [decode(item.encode(SINGLE_BYTE_ENCODING)) for item in lst]


def decode_header(unit, decode):
    """
    The header has been arbitrarily decoded with a single-byte encoding. We
    re-encode it to decode values with the proper encoding defined in the header
    (using decode_list above).
    """
    for attr in (
        "msgctxt",
        "msgid",
        "msgid_pluralcomments",
        "msgid_plural",
        "msgstr",
        "othercomments",
        "automaticcomments",
        "sourcecomments",
        "typecomments",
        "msgidcomments",
    ):
        element = getattr(unit, attr)
        if isinstance(element, list):
            setattr(unit, attr, decode_list(element, decode))
        else:
            setattr(
                unit,
                attr,
                {key: decode_list(value, decode) for key, value in element.items()},
            )


def parse_header(parse_state, store):
    first_unit = parse_unit(parse_state)
    if first_unit is None:
        return None
    set_encoding(parse_state, store, first_unit)
    decode_header(first_unit, parse_state.decode)
    # Fix encoding of next line in parser
    # It was originally parsed with  SINGLE_BYTE_ENCODING
    # but we need to convert it to actual encoding
    parse_state.next_line = parse_state.decode(
        parse_state.next_line.encode(SINGLE_BYTE_ENCODING)
    )
    return first_unit


def parse_units(parse_state, store):
    unit = parse_header(parse_state, store)
    while unit:
        unit.infer_state()
        store.addunit(unit)
        unit = parse_unit(parse_state)
    if not parse_state.eof:
        raise PoParseError(parse_state)
